#!/usr/bin/env python3

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# Copyright 2003-2009 by Paul McGuire
#

OP_REGEX = ["+", "-", "*", "/", "^"]
single_arg_functions = {"exp", "ln", "sin", "cos", "tg", "abs"}
functions = {"sum", "mean", "min", "max", "matmul"}.union(single_arg_functions)

import pyparsing as pp
from pyparsing import ParseException


class Parser(object):
    def __init__(self):
        fnumber = pp.Regex(r"\d+(:?\.\d*)?(:?[eE][+-]?\d+)?")
        identifier = pp.Word(pp.alphas, pp.alphas + pp.nums + "_$")
        func = pp.oneOf(" ".join(functions))
        plus = pp.Literal("+")
        comma = pp.Literal(",")
        minus = pp.Literal("-")
        mul = pp.Literal("*")
        div = pp.Literal("/")
        transpose = pp.Literal(".T")
        lpar = pp.Literal("(").suppress()
        rpar = pp.Literal(")").suppress()
        addop = plus | minus | comma
        mulop = mul | div
        expop = pp.Literal("^") | pp.Literal("**")

        expr = pp.Forward()
        atom = ((0, None) * minus + (fnumber | pp.Group(func + pp.Group(lpar + expr + rpar)) | identifier) |
                pp.Group(lpar + expr + rpar))

        atom = pp.Group(atom + pp.OneOrMore(transpose)) | atom

        factor = pp.Forward()
        factor << atom + pp.ZeroOrMore((expop + factor))
        term = factor + pp.ZeroOrMore((mulop + factor))
        # nextatom = pp.Forward()
        expr << term + pp.ZeroOrMore((addop + expr))
        self.bnf = expr

    def _reformat(self, tokens: list):
        # Handle single token
        while len(tokens) == 1 and isinstance(tokens[0], list):
            tokens = tokens[0]
        if len(tokens) == 1:
            return tokens[0]

        # Handle commas for min/max
        for i, t in enumerate(tokens):
            if t == ",":
                arg0 = self._reformat(tokens[0:i])
                arg1 = self._reformat(tokens[i + 1:])
                return [arg0, arg1]

        # handle addition
        for i, t in enumerate(tokens):
            if t == "+":
                arg0 = self._reformat(tokens[0:i])
                arg1 = self._reformat(tokens[i + 1:])
                return [t, [arg0, arg1]]
        # handle subtraction
        # minus should be interpreted from right to left - thus reversed
        for i, t in reversed(list(enumerate(tokens))):
            if t == "-":
                if i > 0:
                    arg0 = self._reformat(tokens[0:i])
                    arg1 = self._reformat(tokens[i + 1:])
                    return [t, [arg0, arg1]]
                else:
                    return [t, [self._reformat(tokens[1:])]]
        # handle mutlitplication and division
        for i, t in reversed(list(enumerate(tokens))):
            if t in ["*", "/"]:
                arg0 = self._reformat(tokens[0:i])
                arg1 = self._reformat(tokens[i + 1:])
                return [t, [arg0, arg1]]

        for i, t in reversed(list(enumerate(tokens))):
            if t in ["^", "**"]:
                arg0 = self._reformat(tokens[0:i])
                arg1 = self._reformat(tokens[i + 1:])

                return [t, [arg0, arg1]]

        # handle transposition
        if tokens[-1] == ".T":
            assert len(tokens) >= 2
            arg = self._reformat(tokens[0:-1])
            return [".T", [arg]]

        # handle named functions
        if isinstance(tokens[0], str) and tokens[0] in functions:
            assert len(tokens) == 2
            args = self._reformat(tokens[1])
            if tokens[0] in single_arg_functions:
                args = [args]
            return [tokens[0], args]

        raise Exception("Input string is wrong or this method is messed up.\n"
                        "tokens: '{}'".format(tokens))

    def parse(self, s: str):
        raw_parse_result = self.bnf().parseString(s, parseAll=True).asList()
        parse_result = self._reformat(raw_parse_result)
        return parse_result

    bnf = None


class SampleNode(object):
    def __init__(self, value, children=None):
        self.value = value
        self.children = children

    def print(self, depth=0):
        print("■" * depth + str(self.value))
        if self.children is not None:
            for c in self.children:
                c.print(depth + 1)


def build_sample_tree(func):
    if func is None or len(func) == 0:
        raise ValueError
    while isinstance(func, list) and len(func) == 1:
        func = func[0]
    if isinstance(func,list):
        parent = func[0]
    else:
        parent = func
    if not isinstance(func, list):
        children = None
    else:
        children = [build_sample_tree(x) for x in func[1]]

    return SampleNode(parent, children)


if __name__ == "__main__":
    parser = Parser()


    def test(s, ):
        try:
            results = parser.parse(s)
            tree = build_sample_tree(results)
            print(s, "->", results)
            tree.print()
        except ParseException as pe:
            print(s, "failed parse:", str(pe))


    test("x.T +y.T.T +(-a)")
    test("sin(x+y+2).T")
    test("max(x,max(z,y))")
